package lib;

import haxe.ds.IntMap;

/**
	Map values to positions in 3D space. For representing blocks from a scan see `BlockBlockMap`.

	The only reason this exsists is that the abstract class for BlockPos does not
	satisfy the constraint of HashMap. The hashCode function needs to be on a real class and not on 
	an abstract class. Thanks Obama.
**/
@:forward(iterator, clear)
abstract BlockMap<T>(IntMap<T>) {
	/**
		Creates a new HashMap.
	**/
	public inline function new() {
		this = new IntMap();
	}

	/**
		See `Map.set`
	**/
	@:arrayAccess public inline function set(k:BlockPos, v:T) {
		this.set(k.hashCode(), v);
	}

	/**
		See `Map.get`
	**/
	@:arrayAccess public inline function get(k:BlockPos) {
		return this.get(k.hashCode());
	}

	/**
		See `Map.exists`
	**/
	public inline function exists(k:BlockPos) {
		return this.exists(k.hashCode());
	}

	/**
		See `Map.remove`
	**/
	public inline function remove(k:BlockPos) {
		return this.remove(k.hashCode());
	}
}
